define(['jquery', 'lib/gl_lib', 'lib/gl_utils', 'app/binaryFileLoader', 'app/engine/HLBSP', 'app/gui/show_fps'], function($, gl_lib, gl_utils, BinaryFileLoader, HLBSP, ShowFps) {

	var vec3  = gl_lib.vec3;
	var mat3  = gl_lib.mat3;
	var mat4  = gl_lib.mat4;
	var quat4 = gl_lib.quat4;

	var getAvailableContext = gl_utils.getAvailableContext;
	var getShader           = gl_utils.getShader;
	var createShaderProgram = gl_utils.createShaderProgram;
	var createSolidTexture  = gl_utils.createSolidTexture;
	var loadTexture         = gl_utils.loadTexture;

	/*
	 * Copyright (c) 2012 Michael Domanski
	 *
	 * This software is provided 'as-is', without any express or implied
	 * warranty. In no event will the authors be held liable for any damages
	 * arising from the use of this software.
	 *
	 * Permission is granted to anyone to use this software for any purpose,
	 * including commercial applications, and to alter it and redistribute it
	 * freely, subject to the following restrictions:
	 *
	 *    1. The origin of this software must not be misrepresented; you must not
	 *    claim that you wrote the original software. If you use this software
	 *    in a product, an acknowledgment in the product documentation would be
	 *    appreciated but is not required.
	 *
	 *    2. Altered source versions must be plainly marked as such, and must not
	 *    be misrepresented as being the original software.
	 *
	 *    3. This notice may not be removed or altered from any source
	 *    distribution.
	 */

	// Globals
	var Width 	= 800;
	var Height 	= 600;


	/**
	 * Main game object
	 */
	var Game = function(maps, baseUrl) {
		var self = this;
		
		this.modelViewMat = mat4.create();
		this.projectionMat = mat4.create();
		this.camera = mat4.create();
		this.modelViewInvMat = mat3.create();
		this.cameraMat = mat4.create();

		this.cameraPosition = [0, 0, 0];
		this.speed = 200;
		this.zAngle = 3;
		this.xAngle = 0;

		this.frame = 0;
		this.pressed = new Array(128);

		this.activeShader = null;
		this.map = null;

		this.gl = null;
		
		this.time_update_last = 0;
		
		this.load_map_timeout = 5;
		
		this.setMaps(maps);
		
		this.binaryFileLoader = new BinaryFileLoader({
			baseUrl: baseUrl,
			transfer: {
				'de_dust2.bsp': 'de_dust2.bsp',
				'de_dust.bsp': 'de_dust.bsp',
				'fy_snow.bsp': 'fy_snow.bsp',
				
				'de_pripyat.bsp': 'de_pripyat.bsp',
				'pve_neonlines.bsp': 'pve_neonlines.bsp',
				'c1a1f.bsp': 'c1a1f.bsp',
				'fy_snow.bsp': 'fy_snow.bsp',
				
				'decals.wad': 'decals.wad',
				'halflife.wad': 'halflife.wad',
				'sample.wad': 'sample.wad',
				'xeno.wad': 'xeno.wad',
				
				'cs_dust.wad': 'cs_dust.wad',
			}
		});
		
		this.binaryFileLoader.on('loadstart', function(msg) {
			self.addProgressBar(msg.id, msg.filename);
		});
		this.binaryFileLoader.on('progress', function(msg) {
			self.updateProgressBar(msg.id, msg.loaded, msg.total);
		});
		this.binaryFileLoader.on('load', function(msg) {
			self.changeStatusProgressBar(msg.id, 'done');
			//self.removeProgressBar(msg.id, msg.filename);
		});
		this.binaryFileLoader.on('error', function(msg) {
			self.changeStatusProgressBar(msg.id, 'fail');
			//self.removeProgressBar(msg.id, msg.filename);
		});
		
		this.timeline = [0,0,0,];

		this.fov = 90;

		this.gl_width  = 800;
		this.gl_height = 600;
		
		this.real_width  = 800;
		this.real_height = 600;
		
		this.load_list = {__proto__: null};
		
		this.gui = {
			showFps: new ShowFps(1, $('#container'))
		};
		
		this.init();
		
		this.isGame = true;
		
	}

	Game.prototype = {
		setMaps: function(maps) {
			this.maps = maps;
			for(var i in this.maps) {
				this.maps[i].index = Number(i);
			}
		},
		
		init: function() {
			// initialize canvas
			var canvas = this.canvas = document.getElementById("viewport");
			var overlay = document.getElementById("viewport-overlay");
			overlay.style.width = Width + "px";
			overlay.style.height = Height + "px";
			//canvas.width = Width;
			//canvas.height = Height;
			canvas.width = $(canvas).width();
			canvas.height = $(canvas).height();
			//canvas.height = Height;

			this.supportRequestPointerLock = (!!$('#viewport')[0].requestPointerLock || !!$('#viewport')[0].mozRequestPointerLock);
			
			this.gl = getAvailableContext(canvas, ["webgl", "experimental-webgl"]);

			if (!this.gl) {
				alert("Error while obtaining WebGL Context");
			} else {
				this.initEvents();

				this.gl.viewportWidth = canvas.width;
				this.gl.viewportHeight = canvas.height;

				this.initGL();
		
				var self = this;
				var winResizeId = 0;
				$(window).on('resize', function() {
					//$(self.canvas).hide();
					
					(function(_winResizeId) {
						setTimeout(function() {
							if ( _winResizeId !== winResizeId ) { return; }
							
							//$(self.canvas).show();
							self.glUpdate();
						}, 1e3);			
					})(++winResizeId);		
				});

				var self = this;

				// Draw Frames in quick succession
				
				(function() {
					self.drawFrame(self.activeShader);
					self.frame++;
					
					requestAnimationFrame(arguments.callee);
				})();
			}
		},

		loadMap: function(mapIndex, doneload, failload) {
			var self = this;
			
			this.con_show();
			
			this.load_list = {__proto__: null};

			var len = this.maps.length;
			mapIndex = mapIndex = ( len + mapIndex % len ) % len;
			var map = this.maps[mapIndex];
			if ( map ) {
				this.binaryFileLoader.config(map.config);
				if ( !map.config.baseUrl ) {
					this.binaryFileLoader.config({baseUrl: this.baseUrl});
				}				
				
				this._loadMap(map.name, doneload, failload);

				return map;
			}
			
			failload();
		},
		
		_loadMap: function(mapName, doneload, failload) {
			var self = this;

			var __clearIntervalId = setInterval(function() {
				if ( __clearIntervalId === false ) {
					return;
				}
				
				if ( (performance.now() * 1e-3) > self.time_update_last + self.load_map_timeout ) {
					self.error('Bsp file', 'Map "'+mapName+'" parse timeout error');
					failload&&failload();
					delClearInterval();
				}
			}, 1e3);
			
			function delClearInterval() {
				if ( __clearIntervalId !== false ) {
					clearInterval(__clearIntervalId);
					__clearIntervalId = false;
				}			
			}
			
			try {
				
				this.map = new HLBSP(mapName, this, this.gl, this.binaryFileLoader, function() {
						if ( self.isGame ) {
							self.con_hide();
						}
			
						(__clearIntervalId!==false)&&doneload&&doneload();
						delClearInterval();
				}, function() {
						(__clearIntervalId!==false)&&failload&&failload();
						delClearInterval();
				});
			
			} catch(e) {
				delClearInterval();
				(__clearIntervalId!==false)&&failload&&failload();
			}
			
		},
		
		time_update: function() {
			this.time_update_last = performance.now() * 1e-3;
		},
		
		con_is_need_scroll: function() {this.time_update();
			var elem = $('.load-process-info-notice-list')[0];
			if ( elem ) {
				if ( typeof(elem.scrollTop) === 'number' ) {
					var _scroll = elem.scrollTop;
					elem.scrollTop = _scroll + 1;
					if ( elem.scrollTop === _scroll ) {
						return true;
					}
					elem.scrollTop = _scroll;
				}
			}
			return false;
		},
		con_scroll: function(callback) {this.time_update();
			var isScroll = this.con_is_need_scroll();
			
			callback.apply(this);
			
			$('.load-process-info-notice-list').scrollTop(1e9);
		},
		con_clear: function() {this.time_update();
			$('.load-process-info-notice-list ul').empty();
			$('.progress-bar-list ul').empty();	
		},
		notice: function(title, text) {this.time_update();
			this.con_scroll(function() {
				var item = $('<li class="load-process-info-item load-process-info-item-notice" ><div class="load-process-info-title" ></div><div class="load-process-info-text" ><pre></pre></div></li>');
				item.find('.load-process-info-title').text(title);
				item.find('.load-process-info-text pre').text(text);
				$('.load-process-info-notice-list ul').append(item);
				
				( title === this.con_last_title ) && item.find('.load-process-info-title').remove();
				
				this.con_last_title = title;
			});
		},
		warning: function(title, text) {this.time_update();
			this.con_scroll(function() {
				var item = $('<li class="load-process-info-item load-process-info-item-warning" ><div class="load-process-info-title" ></div><div class="load-process-info-text" ><pre></pre></div></li>');
				item.find('.load-process-info-title').text(title);
				item.find('.load-process-info-text pre').text(text);
				$('.load-process-info-notice-list ul').append(item);
				
				( title === this.last_title ) && item.find('.load-process-info-title').remove();
				
				this.last_title = title;
			});
		},
		error: function(title, text) {this.time_update();
			this.con_scroll(function() {
				var item = $('<li class="load-process-info-item load-process-info-item-error" ><div class="load-process-info-title" ></div><div class="load-process-info-text" ><pre></pre></div></li>');
				item.find('.load-process-info-title').text(title);
				item.find('.load-process-info-text pre').text(text);
				$('.load-process-info-notice-list ul').append(item);
				
				( title === this.last_title ) && item.find('.load-process-info-title').remove();
				
				this.last_title = title;
			});
		},
		con_show: function() {this.time_update();
			$('.load-process-info').slideDown();
		},
		con_hide: function() {this.time_update();
			$('.load-process-info').slideUp();
		},
		
		
		createProgressBar: function(title, name) {this.time_update();
			var temp = '<li class="progress-bar-item">\
				<div class="progress-bar-bar">\
					<div class="progress-bar-title"><span class="progress-bar-title-title">Net load</span>&nbsp;&nbsp;&nbsp;&nbsp;<span class="progress-bar-title-filename">de_dust2.bsp</span></div>\
					<div class="progress-bar-text"><span class="progress-bar-current">Unknown</span>&nbsp;/&nbsp;<span class="progress-bar-total">Unknown</span></div>\
					<hr class="progress-bar-bar-inside" style="width: 0%;" >\
				</div>\
			</li>';
			
			var item = $(temp);

			var bar = {
				setTitle: function(text) {
					item.find('.progress-bar-title-title').text(text);
				},
				setName: function(text) {
					item.find('.progress-bar-title-filename').text(text);
				},
				setState: function(current, total) {
					item.find('.progress-bar-current').text(current);
					item.find('.progress-bar-total').text(total);
					item.find('.progress-bar-bar-inside').css({width: '0%'});
					
					if ( typeof(current) === 'string' ) {
						item.find('.progress-bar-bar-inside').css({width: (100).toFixed(0)+'%'});
					} else {
						item.find('.progress-bar-bar-inside').css({width: (current/total*100).toFixed(0)+'%'});
					}
				},
				setStatusDone: function() {
					item.addClass('progress-bar-status-done');
				},
				setStatusFail: function() {
					item.addClass('progress-bar-status-fail');
				},
				
				
				insert: function() {
					$('.progress-bar-list ul').prepend(item);
				},
				remove: function() {
					item.remove();
				},
			};
			
			bar.setTitle(title);
			bar.setName(name);
			bar.setState(0,100);
			
			return bar;
		},
		addProgressBar: function(id, filename) {this.time_update();
			var temp = '<li class="progress-bar-item">\
				<div class="progress-bar-bar">\
					<div class="progress-bar-title"><span class="progress-bar-title-title">Net load</span>&nbsp;&nbsp;&nbsp;&nbsp;<span class="progress-bar-title-filename">de_dust2.bsp</span></div>\
					<div class="progress-bar-text">Unknown&nbsp;/&nbsp;Unknown</div>\
					<hr class="progress-bar-bar-inside" >\
				</div>\
			</li>';
			var item = $(temp);
			
			item.find('.progress-bar-title-filename').html(filename);
			item.find('.progress-bar-text').html('Unknown&nbsp;/&nbsp;Unknown');
			item.find('.progress-bar-bar-inside').css({width: '0%'});
			item.attr('progress-bar-id', id);
			$('.progress-bar-list ul').prepend(item);
		},
		updateProgressBar: function(id, loaded, total) {this.time_update();
			var progressBar = $('[progress-bar-id='+id+']');
			if ( progressBar.length ) {
				progressBar.find('.progress-bar-text').html(loaded + '&nbsp;/&nbsp;' + total);
				progressBar.find('.progress-bar-bar-inside').css({width: (loaded/total*100).toFixed(0)+'%'});
				if ( typeof(loaded) === 'string' ) {
					progressBar.find('.progress-bar-bar-inside').css({width: (100).toFixed(0)+'%'});
				}
			}
		},
		removeProgressBar: function(id, filename) {this.time_update();
			var progressBar = $('[progress-bar-id='+id+']');
			if ( progressBar.length ) {
				progressBar.fadeOut(function() {
					progressBar.remove();
				});
			}
		},
		changeStatusProgressBar: function(id, status) {this.time_update();
			var progressBar = $('[progress-bar-id='+id+']');
			if ( progressBar.length ) {
				switch(status) {
					case 'done':
						progressBar.addClass('progress-bar-status-done');
						break;
					case 'fail':
						progressBar.addClass('progress-bar-status-fail');
						break;
				}
			}
		},
		
		playerStart: function(origin, angles) {
			var game = this;
			function parseVector3(text, def) {
				def = def||[0,0,0];
				if ( typeof(text) !== 'string' ) {
					return def;
				}
				
				var m = text.match(/[^\s]+/g);
				if ( !m || m.length !== 3 ) {
					return def;
				}
				
				for(var i in m) {
					m[i] = parseFloat(m[i]);
					if ( isNaN(m[i]) ) {
						return def;
					}
				}
				
				return [m[0], m[1], m[2]];
			}
			


			var origin = parseVector3( origin );
			var angles = parseVector3( angles );

			game.cameraPosition[0] = -parseFloat(origin[0]);
			game.cameraPosition[1] = -parseFloat(origin[1]);
			game.cameraPosition[2] = -parseFloat(origin[2]);
			
			game.zAngle = ((-parseFloat(angles[1])) * Math.PI/180) || 0;		
		},

		initGL: function() {
			this.gl.clearColor(177/255,217/255,249/255,1);
			this.gl.clearDepth(1);
			this.gl.enable(this.gl.DEPTH_TEST);
			this.gl.depthFunc(this.gl.LEQUAL);
			this.gl.enable(this.gl.BLEND);
			this.gl.blendFunc(this.gl.SRC_ALPHA, this.gl.ONE_MINUS_SRC_ALPHA);

			this.gl.enable(this.gl.CULL_FACE);
			this.gl.cullFace(this.gl.FRONT);

			// setup matrices
			mat4.perspective(this.fov, this.gl.viewportWidth/this.gl.viewportHeight, 1.0, 8192, this.projectionMat);
			mat4.identity(this.camera);

			this.glUpdate();

			this.initShaders(this.gl);
		},
		
		glUpdate: function() {
			this.real_width  = Number($('#container').width() );
			this.real_height = Number($('#container').height());
	
			$(this.canvas).width ( this.real_width  );
			$(this.canvas).height( this.real_height );

			this.canvas.width  = this.gl_width;
			this.canvas.height = this.gl_height;		
			
			this.gl.viewport(0, 0, this.gl_width, this.gl_height);
			mat4.perspective(this.fov, this.real_width/this.real_height, 1.0, 8192, this.projectionMat);
		},

		initShaders: function() {
			this.activeShader = createShaderProgram(this.gl, 'lightmap-vs', 'lightmap-fs');
		},

		drawOrigin: function() {
				// This is our first person movement code. It's not really pretty, but it works
				var self = this;
				
				var currDist = this.speed * self.timeline[2];
				
				var dir = [0, 0, 0];
				if(self.pressed['W'.charCodeAt(0)]) {
					dir[2] += currDist;
				}
				if(self.pressed['S'.charCodeAt(0)]) {
					dir[2] -= currDist;
				}
				if(self.pressed['A'.charCodeAt(0)]) {
					dir[0] += currDist;
				}
				if(self.pressed['D'.charCodeAt(0)]) {
					dir[0] -= currDist;
				}
				if(self.pressed[17]) {
					dir[1] += currDist;
				}
				if(self.pressed[32]) {
					dir[1] -= currDist;
				}

				mat4.identity(self.cameraMat);
				mat4.rotateX(self.cameraMat, self.xAngle-Math.PI/2);
				mat4.rotateZ(self.cameraMat, self.zAngle);
				mat4.inverse(self.cameraMat);

				mat4.multiplyVec3(self.cameraMat, dir);
				vec3.add(self.cameraPosition, dir);
		},
		
		drawFrame: function(shader) {
			if ( !this.isGame ) {
				return;
			}

			this.timeline[1] = this.timeline[0];
			this.timeline[0] = performance.now()*1e-3;
			this.timeline[2] = this.timeline[0] - this.timeline[1];
			if ( this.timeline[1] === 0 ) {
				return;
			}
			
			this.gui.showFps.frame();
			
			this.drawOrigin();
			
			this.gl.clear(this.gl.COLOR_BUFFER_BIT | this.gl.DEPTH_BUFFER_BIT);

			// set default shader
			this.gl.useProgram(shader);



			// setup matrices
			mat4.identity(this.modelViewMat);
			mat4.rotateX(this.modelViewMat, this.xAngle-Math.PI/2);
			mat4.rotateZ(this.modelViewMat, this.zAngle);
			mat4.translate(this.modelViewMat, this.cameraPosition);

			this.gl.uniformMatrix4fv(shader.uniform.modelViewMat , false, this.modelViewMat );
			this.gl.uniformMatrix4fv(shader.uniform.projectionMat, false, this.projectionMat);

			if ( this.map ) {
				this.map.draw(shader);
				this.map.drawEntities(shader);
			}
		},

		initEvents: function() {
			var movingModel = false;
			var lastX = 0;
			var lastY = 0;
			var self = this;

			$(window).keydown(function(event) {
				self.pressed[event.keyCode] = true;
				event.preventDefault();
				return false;
			});

			$(window).keyup(function(event) {
				self.pressed[event.keyCode] = false;
				event.preventDefault();
				return false;
			});

			var baseScreenX = 0;
			var baseScreenY = 0;
			
			$('#viewport').on('click', function(event) {
				
				if ( self.supportRequestPointerLock ) {

					if ( !document.pointerLockElement && !document.mozPointerLockElement ) {
						var viewport = $('#viewport')[0];
						viewport.requestPointerLock    && viewport.requestPointerLock();
						viewport.mozRequestPointerLock && viewport.mozRequestPointerLock();
					} else {
						document.exitPointerLock    && document.exitPointerLock();
						document.mozExitPointerLock && document.mozExitPointerLock();
					}
					
				} else {
				
					if(event.which == 1) {
						movingModel = !movingModel;
					}
				
				}
					
				baseScreenX = event.screenX || 0;
				baseScreenY = event.screenY || 0;
			});
			
			document.addEventListener('pointerlockchange'   , function(e) { movingModel = !!document.pointerLockElement   ; }, false);
			document.addEventListener('mozpointerlockchange', function(e) { movingModel = !!document.mozPointerLockElement; }, false);		

			window.addEventListener('mousemove', function(event) {
				if ( 'movementY' in event ) {
					var xDelta = event.movementX;
					var yDelta = event.movementY;
				} else {
					var xDelta = event.screenX - baseScreenX;
					var yDelta = event.screenY - baseScreenY;
					baseScreenX = event.screenX;
					baseScreenY = event.screenY;
				}			

				var piDiv2 = Math.PI / 2;
				var piMul2 = Math.PI * 2;
					
				function normalizeAngle(angle) {
					if ( angle > piMul2 ) {
						angle -= piMul2;
					}
					if ( angle < -piMul2 ) {
						angle += piMul2;
					}
					
					return angle;
				}
				
				if ( movingModel ) {
					self.zAngle += xDelta*0.025;
					self.xAngle += yDelta*0.025;
					
					self.zAngle = normalizeAngle(self.zAngle);
					self.xAngle = normalizeAngle(self.xAngle);

					if ( self.xAngle < -piDiv2 ) { self.xAngle = -piDiv2; }
					if ( self.xAngle >  piDiv2 ) { self.xAngle =  piDiv2; }
				}
			});
		}
	}

	return Game;
});